@@ -1,33 +1,34 @@
-# $Id: Code.pm,v 1.31 2004/02/25 23:03:10 jeff Exp $
+# $Id: Code.pm,v 1.36 2004/04/12 15:58:47 jeff Exp $
 
 package ExtProc::Code;
 
 use 5.6.1;
 use strict;
 use warnings;
-no strict qw(subs);
+use AutoLoader 'AUTOLOAD';
 
 require Exporter;
-require DynaLoader;
 
-our @ISA = qw(Exporter DynaLoader);
+our @ISA = qw(Exporter);
 our %EXPORT_TAGS = ( 'all' => [ qw(
 	&create_wrapper
-        &import_code
-        &drop_code
+	&import_code
+	&drop_code
 ) ] );
 our @EXPORT_OK = ( @{ $EXPORT_TAGS{'all'} } );
 our @EXPORT = qw(
 );
-our $VERSION = '1.99_07';
+our $VERSION = '1.99_08';
 
-use ExtProc;
+use ExtProc qw(ep_debug put_line);
 use File::Spec;
 use DBI;
 
+use vars qw(%typemap $c_prefix);
+
 # for converting from *supported* PL/SQL datatypes to C datatypes
 # should explore integrating this with OTT somehow
-my %typemap = (
+%typemap = (
 	'PLS_INTEGER' => {
 		'IN'		=> 'int',
 		'OUT'		=> 'int *',
@@ -71,7 +72,11 @@ my %typemap = (
 );
 
 # prefix for C functions to avoid name clashes with standard C library
-our $c_prefix = "EP_";
+$c_prefix = "EP_";
+
+__END__
+
+### AUTOLOADED SUBROUTINES ###
 
 # create_wrapper(proto,[lib])
 # create C wrapper function in trusted code directory based on prototype
@@ -84,87 +89,117 @@ sub create_wrapper
 	# @args = ( { spec => x, type => x, inout => x, carg => x } )
 	my @args;
 
+	# stuff extracted from prototype
+	my ($subtype, $name, $argstr, $retstr, $rettype);
+
 	# if prototype is valid, write out C wrapper to trusted directory
-	# prototype format: name([arg1[,arg2,...]]) [RETURN type]
-	if ($proto =~ /^(\w+)\s*\((.*?)\)(?:\s+RETURN\s+([\w*]+))*$/io) {
-		my $name = $1;
-		my @a = split(/,\s*/, $2);
-		my $rettype = (defined $3) ? uc($3) : 'void';
-		my $subtype = (defined $3) ?
-			'EP_SUBTYPE_FUNCTION' :
-			'EP_SUBTYPE_PROCEDURE';
-		my $n = 0;
+	# prototype format: SUBTYPE name([arg1[,arg2,...]]) [RETURN type]
 
-		# convert prototype to C arguments
-		foreach (@a) {
-			my ($argname, $inout, $type);
-			if (/^([\w_]+)\s+IN\s+OUT\s+(.+)$/i) {
-				$argname = $1;
-				$inout = "IN OUT";
-				$type = $2;
-			}
-			else {
-				($argname, $inout, $type) = split(/\s+/);
-			}
-			$type = uc($type);
-			unless (exists $typemap{$type}) {
-				die "unsupported datatype: $type";
-			}
-			$inout = uc($inout);
-			my $ctype = $typemap{$type}{$inout};
-			$args[$n]{'spec'} = $_;
-			$args[$n]{'name'} = $argname;
-			$args[$n]{'type'} = $type;
-			$args[$n]{'inout'} = $inout;
-			my $tmp = "$ctype arg$n";
-			if ($typemap{$type}{'NULLABLE'}) {
-				if ($inout =~ /OUT/) {
-					$tmp .= ", OCIInd *is_null_$n";
-				}
-				else {
-					$tmp .= ", OCIInd is_null_$n";
-				}
-			}
-			if ($typemap{$type}{'VARLENGTH'}) {
-				if ($inout =~ /OUT/) {
-					$tmp .= ", sb4 *length_$n, sb4 *maxlen_$n";
-				}
-				else {
-					$tmp .= ", sb4 length_$n";
-				}
-			}
-			$args[$n]{'carg'} = $tmp;
-			$n++;
-		}
-		my $argstr = join(', ', map($_->{'carg'}, @args));
-		$argstr = ", $argstr" if $n;
+	# sub with arguments
+	if ($proto =~ /^(FUNCTION|PROCEDURE)\s+([\w\d_\-]+)\(([^\}]+)\)(.*)$/oi) {
+		$subtype = "EP_SUBTYPE_".uc($1);
+		$name = $2;
+		$argstr = $3;
+		$retstr = $4;
+	}
+	# sub with no arguments
+	elsif ($proto =~ /^(FUNCTION|PROCEDURE)\s+([\w\d_\-]+)(?:\s+(.+))*$/oi) {
+		$subtype = "EP_SUBTYPE_".uc($1);
+		$name = $2;
+		$retstr = $3;
+	}
+	else {
+		die "invalid prototype";
+	}
 
-		# result declaration and fatal return statement
-		my $return_fatal;
-		unless (exists $typemap{$rettype} || $rettype eq 'void') {
-			die "unsupported datatype: $rettype";
+	# get return type, if any
+	if ($retstr =~ /return\s+([\w\d_\-]+)/i) {
+        	$rettype = uc($1);
+	}
+	else {
+		if ($subtype eq "EP_SUBTYPE_FUNCTION") {
+			die "function has no return type";
 		}
-		my $crettype = $typemap{$rettype}{'RETURN'};
-		my $result_dec = "$crettype res;";
-		if ($crettype eq 'void') {
-			$return_fatal = "return";
-			$result_dec = "";
-		}
-		elsif ($crettype eq 'int' || $crettype eq 'float') {
-			$return_fatal = "return(0)";
+		$rettype = 'void';
+	}
+
+	my @a = split(/,\s*/, $argstr);
+	my $n = 0;
+
+	# convert prototype to C arguments
+	foreach (@a) {
+		my ($argname, $inout, $type);
+		if (/^([\w_]+)\s+IN\s+OUT\s+(.+)$/i) {
+			$argname = $1;
+			$inout = "IN OUT";
+			$type = $2;
 		}
 		else {
-			$return_fatal = "return(NULL)";
+			($argname, $inout, $type) = split(/\s+/);
 		}
+		$type = uc($type);
+		unless (exists $typemap{$type}) {
+			die "unsupported datatype: $type";
+		}
+		$inout = uc($inout);
+		my $ctype = $typemap{$type}{$inout};
+		$args[$n]{'spec'} = $_;
+		$args[$n]{'name'} = $argname;
+		$args[$n]{'type'} = $type;
+		$args[$n]{'inout'} = $inout;
+		my $tmp = "$ctype arg$n";
+		if ($typemap{$type}{'NULLABLE'}) {
+			if ($inout =~ /OUT/) {
+				$tmp .= ", OCIInd *is_null_$n";
+			}
+			else {
+				$tmp .= ", OCIInd is_null_$n";
+			}
+		}
+		if ($typemap{$type}{'VARLENGTH'}) {
+			if ($inout =~ /OUT/) {
+				$tmp .= ", sb4 *length_$n, sb4 *maxlen_$n";
+			}
+			else {
+				$tmp .= ", sb4 length_$n";
+			}
+		}
+		$args[$n]{'carg'} = $tmp;
+		$n++;
+	}
+	my $cargstr = join(', ', map($_->{'carg'}, @args));
+	$cargstr = ", $cargstr" if $n;
 
-		# write to file in trusted code directory
-		local *CODE;
-		my $dir = ExtProc::config('trusted_code_directory');
-		open(CODE, '>', File::Spec->catfile($dir, $name . '.c' ))
-			or die $!;
+	# result declaration and fatal return statement
+	my $return_fatal;
+	unless (exists $typemap{$rettype} || $rettype eq 'void') {
+		die "unsupported return type: $rettype";
+	}
+	my $crettype = $typemap{$rettype}{'RETURN'};
+	my $result_dec = "$crettype res;";
+	if ($crettype eq 'void') {
+		$return_fatal = "return";
+		$result_dec = "";
+	}
+	elsif ($crettype eq 'int' || $crettype eq 'float') {
+		$return_fatal = "return(0)";
+	}
+	else {
+		$return_fatal = "return(NULL)";
+	}
 
-		# generate C wrapper source file
-		print CODE <<_DONE_;
+	# write to file in trusted code directory
+	local *CODE;
+	my $dir = ExtProc::config('trusted_code_directory');
+	open(CODE, '>', File::Spec->catfile($dir, $name . '.c' ))
+		or die $!;
+
+#############################################################################
+# here lies ugliness -- start of C code
+#############################################################################
+
+	# generate C wrapper source file
+	print CODE <<_DONE_;
 /* THIS FILE IS AUTOGENERATED -- CHANGES MAY BE LOST */
 
 #ifdef __cplusplus
@@ -193,15 +228,15 @@ extern EP_CONTEXT my_context;
 
 _DONE_
 
-		# don't need return indicator for procedures
-		if ($rettype eq 'void') {
-			print CODE "$crettype $c_prefix$name(OCIExtProcContext *ctx $argstr)\n";
-		}
-		else {
-			print CODE "$crettype $c_prefix$name(OCIExtProcContext *ctx, OCIInd *ret_ind $argstr)\n";
-		}
+	# don't need return indicator for procedures
+	if ($rettype eq 'void') {
+		print CODE "$crettype $c_prefix$name(OCIExtProcContext *ctx $cargstr)\n";
+	}
+	else {
+		print CODE "$crettype $c_prefix$name(OCIExtProcContext *ctx, OCIInd *ret_ind $cargstr)\n";
+	}
 
-		print CODE <<_DONE_;
+	print CODE <<_DONE_;
 {
 	int nret;
 	short ind;
@@ -221,14 +256,14 @@ _DONE_
 	_ep_init(c, ctx);
 _DONE_
 
-		if ($rettype eq 'void') {
-			print CODE "\tEP_DEBUGF(c, \"IN (user defined) $c_prefix$name(%p, ...)\", ctx);\n";
-		}
-		else {
-			print CODE "\tEP_DEBUGF(c, \"IN (user defined) $c_prefix$name(%p, %p, ...)\", ctx, ret_ind);\n";
-		}
+	if ($rettype eq 'void') {
+		print CODE "\tEP_DEBUGF(c, \"IN (user defined) $c_prefix$name(%p, ...)\", ctx);\n";
+	}
+	else {
+		print CODE "\tEP_DEBUGF(c, \"IN (user defined) $c_prefix$name(%p, %p, ...)\", ctx, ret_ind);\n";
+	}
 
-		print CODE <<_DONE_;
+	print CODE <<_DONE_;
 	EP_DEBUG(c, "-- prototype: $proto");
 
 	c->subtype = $subtype;
@@ -254,11 +289,11 @@ _DONE_
 		if (!c->perl) {
 _DONE_
 
-		if ($rettype ne 'void') {
-			print CODE "\t\t\t*ret_ind = OCI_IND_NULL;\n";
-		}
+	if ($rettype ne 'void') {
+		print CODE "\t\t\t*ret_ind = OCI_IND_NULL;\n";
+	}
 
-		print CODE <<_DONE_;
+	print CODE <<_DONE_;
 			$return_fatal;
 		}
 	}
@@ -270,58 +305,63 @@ _DONE_
 	PUSHMARK(SP);
 
 _DONE_
-		foreach my $n (0..$#args) {
-			print CODE "\t/* push arg$n ($args[$n]{'spec'}) onto stack */\n";
-			if ($typemap{$args[$n]{'type'}}{'NULLABLE'}) {
-				if ($args[$n]{'inout'} eq 'IN') {
-					print CODE <<_DONE_;
+
+#############################################################################
+# convert arguments to perl types and push onto stack
+#############################################################################
+
+	foreach my $n (0..$#args) {
+		print CODE "\t/* push arg$n ($args[$n]{'spec'}) onto stack */\n";
+		if ($typemap{$args[$n]{'type'}}{'NULLABLE'}) {
+			if ($args[$n]{'inout'} eq 'IN') {
+				print CODE <<_DONE_;
 	if (is_null_$n == OCI_IND_NULL) {
 		sv = sv_2mortal(newSVsv(&PL_sv_undef));
 	}
 	else {
 _DONE_
-				}
-				else {
-					if ($args[$n]{'type'} eq 'VARCHAR2') {
-						if ($args[$n]{'inout'} eq 'OUT') {
-							print CODE <<_DONE_;
+			}
+			else {
+				if ($args[$n]{'type'} eq 'VARCHAR2') {
+					if ($args[$n]{'inout'} eq 'OUT') {
+						print CODE <<_DONE_;
 	sv = sv_2mortal(newSVsv(&PL_sv_undef));
 	{ /* placeholder brace */
 _DONE_
-						}
-						else {
-							print CODE <<_DONE_;
+					}
+					else {
+						print CODE <<_DONE_;
 	if (*is_null_$n == OCI_IND_NULL) {	
 		sv = sv_2mortal(newSVsv(&PL_sv_undef));
 	}
 	else {
 _DONE_
-						}
 					}
-					else {
-						print CODE "\t{ /* placeholder brace */\n";
-					}
-				}
-			}
-			my $star = ($args[$n]{'inout'} =~ /OUT/) ? "*" : "";
-			if ($args[$n]{'carg'} =~ /^char /) {
-				if ($args[$n]{'inout'} =~ /IN/) {
-					# for IN modes
-					print CODE "\tsv = sv_2mortal(newSVpvn(arg$n, ${star}length_$n));\n";
 				}
 				else {
-					# for IN OUT & OUT modes
-					print CODE "\tsv = sv_newmortal();\n";
+					print CODE "\t{ /* placeholder brace */\n";
 				}
 			}
-			elsif ($args[$n]{'carg'} =~ /^int /) {
-				print CODE "\tsv = sv_2mortal(newSViv(${star}arg$n));\n";
+		}
+		my $star = ($args[$n]{'inout'} =~ /OUT/) ? "*" : "";
+		if ($args[$n]{'carg'} =~ /^char /) {
+			if ($args[$n]{'inout'} =~ /IN/) {
+				# for IN modes
+				print CODE "\tsv = sv_2mortal(newSVpvn(arg$n, ${star}length_$n));\n";
 			}
-			elsif ($args[$n]{'carg'} =~ /^float /) {
-				print CODE "\tsv = sv_2mortal(newSVnv(${star}arg$n));\n";
+			else {
+				# for IN OUT & OUT modes
+				print CODE "\tsv = sv_newmortal();\n";
 			}
-			elsif ($args[$n]{'carg'} =~ /^OCIDate /) {
-				print CODE <<_DONE_;
+		}
+		elsif ($args[$n]{'carg'} =~ /^int /) {
+			print CODE "\tsv = sv_2mortal(newSViv(${star}arg$n));\n";
+		}
+		elsif ($args[$n]{'carg'} =~ /^float /) {
+			print CODE "\tsv = sv_2mortal(newSVnv(${star}arg$n));\n";
+		}
+		elsif ($args[$n]{'carg'} =~ /^OCIDate /) {
+			print CODE <<_DONE_;
 	sv = sv_newmortal();
 	sv_setref_pv(sv, "ExtProc::DataType::OCIDate", arg$n);
 	if (*is_null_$n == OCI_IND_NULL) {
@@ -331,43 +371,47 @@ _DONE_
 		clear_null(arg$n); /* in case we used this address before */
 	}
 _DONE_
-			}
-			else {
-				die "unsupported C datatype: $args[$n]{'carg'} (was $args[$n]{'spec'})";
-			}
+		}
+		else {
+			die "unsupported C datatype: $args[$n]{'carg'} (was $args[$n]{'spec'})";
+		}
 
-			print CODE <<_DONE_;
+		print CODE <<_DONE_;
 	svcache[$n] = sv;
 	if (c->tainting) {
 		SvTAINTED_on(sv);
 	}
 _DONE_
-			if ($typemap{$args[$n]{'type'}}{'NULLABLE'}) {
-				print CODE "\t}\n";
-			}
-
-			# IN OUT & OUT types are always passed as references
-			# leave SV alone if it's already a reference
-			if ($args[$n]{'inout'} =~ /OUT/) {
-				print CODE "\tXPUSHs(sv_isobject(sv) ? sv : newRV_noinc(sv));\n";
-			}
-			else {
-				print CODE "\tXPUSHs(sv);\n";
-			}
+		if ($typemap{$args[$n]{'type'}}{'NULLABLE'}) {
+			print CODE "\t}\n";
 		}
 
-		print CODE <<_DONE_;
+		# IN OUT & OUT types are always passed as references
+		# leave SV alone if it's already a reference
+		if ($args[$n]{'inout'} =~ /OUT/) {
+			print CODE "\tXPUSHs(sv_isobject(sv) ? sv : newRV_noinc(sv));\n";
+		}
+		else {
+			print CODE "\tXPUSHs(sv);\n";
+		}
+	}
+
+#############################################################################
+# parse and run subroutine
+#############################################################################
+
+	print CODE <<_DONE_;
 	PUTBACK;
 
 	fqsub = parse_code(c, &code, "$name");
 	EP_DEBUG(c, "RETURN (user defined) $c_prefix$name");
 	if (!fqsub) {
 _DONE_
-			if ($rettype ne 'void') {
-				print CODE "\t\t*ret_ind = OCI_IND_NULL;\n";
-			}
+		if ($rettype ne 'void') {
+			print CODE "\t\t*ret_ind = OCI_IND_NULL;\n";
+		}
 
-			print CODE <<_DONE_;
+		print CODE <<_DONE_;
 		ora_exception(c, "invalid subroutine");
 		$return_fatal;
 	}
@@ -378,11 +422,15 @@ _DONE_
 	SPAGAIN;
 _DONE_
 
-		# copy values to IN OUT and OUT args
-		foreach my $n (0..$#args) {
-			if ($args[$n]{'inout'} =~ /OUT/) {
-				if ($args[$n]{'carg'} =~ /^char /) {
-					print CODE <<_DONE_;
+#############################################################################
+# process IN OUT & OUT return values
+#############################################################################
+
+	# copy values to IN OUT and OUT args
+	foreach my $n (0..$#args) {
+		if ($args[$n]{'inout'} =~ /OUT/) {
+			if ($args[$n]{'carg'} =~ /^char /) {
+				print CODE <<_DONE_;
 	if (!SvOK(svcache[$n])) {
 		*is_null_$n = OCI_IND_NULL;
 	}
@@ -398,34 +446,38 @@ _DONE_
 		*length_$n = len;
 	}
 _DONE_
-				}
-				elsif ($args[$n]{'carg'} =~ /^OCIDate /) {
-					print CODE <<_DONE_;
+			}
+			elsif ($args[$n]{'carg'} =~ /^OCIDate /) {
+				print CODE <<_DONE_;
 	/* DATE types are passed as pointers, so no need to copy again */
 	*is_null_$n = is_null(arg$n) ? OCI_IND_NULL : OCI_IND_NOTNULL;
 _DONE_
-				}
-				elsif ($args[$n]{'carg'} =~ /^int /) {
-					print CODE <<_DONE_;
+			}
+			elsif ($args[$n]{'carg'} =~ /^int /) {
+				print CODE <<_DONE_;
 	*arg$n = SvIV(svcache[$n]);
 	*is_null_$n = SvOK(svcache[$n]) ? OCI_IND_NOTNULL : OCI_IND_NULL;
 _DONE_
-				}
-				elsif ($args[$n]{'carg'} =~ /^float /) {
-					print CODE <<_DONE_;
+			}
+			elsif ($args[$n]{'carg'} =~ /^float /) {
+				print CODE <<_DONE_;
 	*arg$n = SvNV(svcache[$n]);
 	*is_null_$n = SvOK(svcache[$n]) ? OCI_IND_NOTNULL : OCI_IND_NULL;
 _DONE_
-				}
-				else {
-					die "unsupported C datatype: $args[$n]{'carg'} (was $args[$n]{'spec'})";
-				}
+			}
+			else {
+				die "unsupported C datatype: $args[$n]{'carg'} (was $args[$n]{'spec'})";
 			}
 		}
+	}
 
-		# procedures don't return values
-		if ($rettype eq 'void') {
-			print CODE <<_DONE_;
+#############################################################################
+# return from a procedure
+#############################################################################
+
+	# procedures don't return values
+	if ($rettype eq 'void') {
+		print CODE <<_DONE_;
 	/* clean up stack and return */
 	PUTBACK;
 	FREETMPS;
@@ -433,16 +485,21 @@ _DONE_
 	return;
 }
 _DONE_
-		}
-		# functions do return values
-		else {
-			print CODE <<_DONE_;
+	}
+
+#############################################################################
+# convert values back to oracle types and return value from a function
+#############################################################################
+
+	# functions do return values
+	else {
+		print CODE <<_DONE_;
 
 	/* grab return value off the stack */
 	sv = POPs;
 _DONE_
-			if ($crettype =~ /char\s*\*/) {
-				print CODE <<_DONE_;
+		if ($crettype =~ /char\s*\*/) {
+			print CODE <<_DONE_;
 	if (SvOK(sv)) {
 		tmp = SvPV(sv,len);
 		New(0, res, len+1, char);
@@ -454,30 +511,30 @@ _DONE_
 		*ret_ind = OCI_IND_NULL;
 	}
 _DONE_
-			}
-			elsif ($crettype =~ /OCIDate\s*\*/) {
-				print CODE <<_DONE_;
+		}
+		elsif ($crettype =~ /OCIDate\s*\*/) {
+			print CODE <<_DONE_;
 	res = ($crettype)SvIV(SvRV(sv));
 	*ret_ind = is_null(sv) ? OCI_IND_NULL : OCI_IND_NOTNULL;
 _DONE_
-			}
-			elsif ($crettype eq 'int') {
-				print CODE <<_DONE_;
+		}
+		elsif ($crettype eq 'int') {
+			print CODE <<_DONE_;
 	res = SvIV(sv);
 	*ret_ind = SvTRUE(sv) ? OCI_IND_NOTNULL : OCI_IND_NULL;
 _DONE_
-			}
-			elsif ($crettype eq 'float') {
-				print CODE <<_DONE_;
+		}
+		elsif ($crettype eq 'float') {
+			print CODE <<_DONE_;
 	res = SvNV(sv);
 	*ret_ind = SvTRUE(sv) ? OCI_IND_NOTNULL : OCI_IND_NULL;
 _DONE_
-			}
-			else {
-				die "unknown return type: $crettype";
-			}
+		}
+		else {
+			die "unknown return type: $crettype";
+		}
 
-			print CODE <<_DONE_;
+		print CODE <<_DONE_;
 
 	/* clean up stack and return */
 	PUTBACK;
@@ -487,43 +544,57 @@ _DONE_
 	return(res);
 }
 _DONE_
-		}
-		close(CODE);
+	}
+	close(CODE);
 
-		# external procedure spec
-		local *SPEC;
-		open(SPEC, '>', File::Spec->catfile($dir, $name . '.sql' ))
-			or die $!;
-		print SPEC "CREATE OR REPLACE ", ($rettype eq 'void') ?
-			"PROCEDURE" : "FUNCTION", " $proto\n";
-		print SPEC "AS EXTERNAL NAME \"EP_$name\"\n";
-		print SPEC "LIBRARY \"", $lib ? $lib : "PERL_LIB", "\"\n";
-		print SPEC "WITH CONTEXT\n";
-		print SPEC "PARAMETERS (\n";
-		print SPEC "   CONTEXT";
-		($rettype eq 'void') or print SPEC ",\n   RETURN INDICATOR BY REFERENCE\n";
-		foreach my $n (0..$#args) {
-			my $name = $args[$n]{'name'};
-			my $type = $args[$n]{'type'};
-			my $inout = $args[$n]{'inout'};
-			print SPEC ",\n   $name $typemap{$type}{'PARAMTYPE'}";
-			if ($typemap{$type}{'NULLABLE'}) {
-				print SPEC ",\n   $name INDICATOR short";
-			}
-			if ($typemap{$type}{'VARLENGTH'}) {
-				print SPEC ",\n   $name LENGTH sb4";
-				if ($inout =~ /OUT/) {
-					print SPEC ",\n   $name MAXLEN sb4";
-				}
-			}
-		}
-		print SPEC "\n);\n/\n";
-		close(SPEC);
+#############################################################################
+# generate and save DDL
+#############################################################################
+
+	# external procedure DDL
+	my $sql;
+	my $ddl_format = ExtProc::config('ddl_format');
+	if ($ddl_format == 0) {
+		$sql = 'CREATE OR REPLACE ';
 	}
 	else {
-		die "invalid prototype";
+		$sql = "-- for package specification\n$proto\n";
+		$sql .= "-- for package body\n";
 	}
+	$sql .= "$proto\n";
+	$sql .= "AS EXTERNAL NAME \"EP_$name\"\n";
+	$sql .= "LIBRARY \"" . ($lib ? $lib : "PERL_LIB") . "\"\n";
+	$sql .= "WITH CONTEXT\n";
+	$sql .= "PARAMETERS (\n";
+	$sql .= "   CONTEXT";
+	$sql .= ",\n   RETURN INDICATOR BY REFERENCE"
+		unless ($rettype eq 'void');
+	foreach my $n (0..$#args) {
+		my $name = $args[$n]{'name'};
+		my $type = $args[$n]{'type'};
+		my $inout = $args[$n]{'inout'};
+		$sql .= ",\n   $name $typemap{$type}{'PARAMTYPE'}";
+		if ($typemap{$type}{'NULLABLE'}) {
+			$sql .= ",\n   $name INDICATOR short";
+		}
+		if ($typemap{$type}{'VARLENGTH'}) {
+			$sql .= ",\n   $name LENGTH sb4";
+			if ($inout =~ /OUT/) {
+				$sql .= ",\n   $name MAXLEN sb4";
+			}
+		}
+	}
+	$sql .= "\n);";
 
+	# write DDL to file
+	local *DDL;
+	open(DDL, '>', File::Spec->catfile($dir, $name . '.sql' ))
+		or die $!;
+	print DDL "$sql\n";
+	close(DDL);
+
+	# output DDL ("set serveroutput on" to see it in sqlplus)
+	put_line($_) foreach (split(/[\r\n]+/, $sql));
 }
 
 # import_code(name, filename, [proto])
@@ -586,15 +657,18 @@ sub import_code
 	$sth->execute($name);
 	$sth->finish;
 
+	# find prototype, if any
+	if (!$proto && $code =~ /^((?:FUNCTION|PROCEDURE)\s+$name.*)/im) {
+		$proto = $1;
+	}
+
 	# import code into database
-	$sth = $dbh->prepare("insert into $table (name, language, last_modified_user, last_modified_date, code) values(?, 'Perl5', user, SYSDATE, ?)");
-	$sth->execute($name, $code);
+	$sth = $dbh->prepare("insert into $table (name, plsql_spec, language, code) values(?, ?, 'Perl5', ?)");
+	$sth->execute($name, $proto, $code);
 	$sth->finish;
 
-	# if we have a prototype, create the wrapper
-	if (defined($proto)) {
-		create_wrapper($proto);
-	}
+	# create C wrapper if we have a prototype
+	$proto && create_wrapper($proto);
 }
 
 # drop_code(name)
